(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(require('tweakpane')) :
    typeof define === 'function' && define.amd ? define(['tweakpane'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Tweakpane));
}(this, (function (Tweakpane) { 'use strict';

    function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

    var Tweakpane__default = /*#__PURE__*/_interopDefaultLegacy(Tweakpane);

    /**
     * A type-safe event emitter.
     * @template E The interface that maps event names and event objects.
     */
    var Emitter = /** @class */ (function () {
        function Emitter() {
            this.observers_ = {};
        }
        /**
         * Adds an event listener to the emitter.
         * @param eventName The event name to listen.
         * @param handler The event handler.
         */
        Emitter.prototype.on = function (eventName, handler) {
            var observers = this.observers_[eventName];
            if (!observers) {
                observers = this.observers_[eventName] = [];
            }
            observers.push({
                handler: handler,
            });
            return this;
        };
        /**
         * Removes an event listener from the emitter.
         * @param eventName The event name.
         * @param handler The event handler to remove.
         */
        Emitter.prototype.off = function (eventName, handler) {
            var observers = this.observers_[eventName];
            if (observers) {
                this.observers_[eventName] = observers.filter(function (observer) {
                    return observer.handler !== handler;
                });
            }
            return this;
        };
        Emitter.prototype.emit = function (eventName, event) {
            var observers = this.observers_[eventName];
            if (!observers) {
                return;
            }
            observers.forEach(function (observer) {
                observer.handler(event);
            });
        };
        return Emitter;
    }());

    var PrimitiveValue = /** @class */ (function () {
        function PrimitiveValue(initialValue) {
            this.emitter = new Emitter();
            this.value_ = initialValue;
        }
        Object.defineProperty(PrimitiveValue.prototype, "rawValue", {
            get: function () {
                return this.value_;
            },
            set: function (value) {
                if (this.value_ === value) {
                    return;
                }
                this.value_ = value;
                this.emitter.emit('change', {
                    sender: this,
                    rawValue: this.value_,
                });
            },
            enumerable: false,
            configurable: true
        });
        return PrimitiveValue;
    }());

    function forceCast(v) {
        return v;
    }
    function isEmpty(value) {
        return value === null || value === undefined;
    }

    var SVG_NS = 'http://www.w3.org/2000/svg';
    function supportsTouch(doc) {
        return doc.ontouchstart !== undefined;
    }
    function removeElement(element) {
        if (element.parentElement) {
            element.parentElement.removeChild(element);
        }
    }

    function computeOffset(ev, elem) {
        // NOTE: OffsetX/Y should be computed from page and window properties to capture mouse events
        var win = elem.ownerDocument.defaultView;
        var rect = elem.getBoundingClientRect();
        return {
            x: ev.pageX - (((win && win.scrollX) || 0) + rect.left),
            y: ev.pageY - (((win && win.scrollY) || 0) + rect.top),
        };
    }
    /**
     * A utility class to handle both mouse and touch events.
     */
    var PointerHandler = /** @class */ (function () {
        function PointerHandler(element) {
            this.onDocumentMouseMove_ = this.onDocumentMouseMove_.bind(this);
            this.onDocumentMouseUp_ = this.onDocumentMouseUp_.bind(this);
            this.onMouseDown_ = this.onMouseDown_.bind(this);
            this.onTouchEnd_ = this.onTouchEnd_.bind(this);
            this.onTouchMove_ = this.onTouchMove_.bind(this);
            this.onTouchStart_ = this.onTouchStart_.bind(this);
            this.elem_ = element;
            this.emitter = new Emitter();
            var doc = this.elem_.ownerDocument;
            if (supportsTouch(doc)) {
                element.addEventListener('touchstart', this.onTouchStart_);
                element.addEventListener('touchmove', this.onTouchMove_);
                element.addEventListener('touchend', this.onTouchEnd_);
            }
            else {
                element.addEventListener('mousedown', this.onMouseDown_);
            }
        }
        PointerHandler.prototype.computePosition_ = function (offset) {
            var rect = this.elem_.getBoundingClientRect();
            return {
                bounds: {
                    width: rect.width,
                    height: rect.height,
                },
                point: offset
                    ? {
                        x: offset.x,
                        y: offset.y,
                    }
                    : null,
            };
        };
        PointerHandler.prototype.onMouseDown_ = function (e) {
            var _a;
            // Prevent native text selection
            e.preventDefault();
            (_a = e.currentTarget) === null || _a === void 0 ? void 0 : _a.focus();
            var doc = this.elem_.ownerDocument;
            doc.addEventListener('mousemove', this.onDocumentMouseMove_);
            doc.addEventListener('mouseup', this.onDocumentMouseUp_);
            this.emitter.emit('down', {
                data: this.computePosition_(computeOffset(e, this.elem_)),
                sender: this,
            });
        };
        PointerHandler.prototype.onDocumentMouseMove_ = function (e) {
            this.emitter.emit('move', {
                data: this.computePosition_(computeOffset(e, this.elem_)),
                sender: this,
            });
        };
        PointerHandler.prototype.onDocumentMouseUp_ = function (e) {
            var doc = this.elem_.ownerDocument;
            doc.removeEventListener('mousemove', this.onDocumentMouseMove_);
            doc.removeEventListener('mouseup', this.onDocumentMouseUp_);
            this.emitter.emit('up', {
                data: this.computePosition_(computeOffset(e, this.elem_)),
                sender: this,
            });
        };
        PointerHandler.prototype.onTouchStart_ = function (e) {
            // Prevent native page scroll
            e.preventDefault();
            var touch = e.targetTouches.item(0);
            var rect = this.elem_.getBoundingClientRect();
            this.emitter.emit('down', {
                data: this.computePosition_(touch
                    ? {
                        x: touch.clientX - rect.left,
                        y: touch.clientY - rect.top,
                    }
                    : undefined),
                sender: this,
            });
        };
        PointerHandler.prototype.onTouchMove_ = function (e) {
            var touch = e.targetTouches.item(0);
            var rect = this.elem_.getBoundingClientRect();
            this.emitter.emit('move', {
                data: this.computePosition_(touch
                    ? {
                        x: touch.clientX - rect.left,
                        y: touch.clientY - rect.top,
                    }
                    : undefined),
                sender: this,
            });
        };
        PointerHandler.prototype.onTouchEnd_ = function (e) {
            var touch = e.targetTouches.item(0);
            var rect = this.elem_.getBoundingClientRect();
            this.emitter.emit('up', {
                data: this.computePosition_(touch
                    ? {
                        x: touch.clientX - rect.left,
                        y: touch.clientY - rect.top,
                    }
                    : undefined),
                sender: this,
            });
        };
        return PointerHandler;
    }());

    function mapRange(value, start1, end1, start2, end2) {
        var p = (value - start1) / (end1 - start1);
        return start2 + p * (end2 - start2);
    }
    function getDecimalDigits(value) {
        var text = String(value.toFixed(10));
        var frac = text.split('.')[1];
        return frac.replace(/0+$/, '').length;
    }
    function constrainRange(value, min, max) {
        return Math.min(Math.max(value, min), max);
    }

    var PREFIX = 'tp';
    /**
     * A utility function for generating BEM-like class name.
     * @param viewName The name of the view. Used as part of the block name.
     * @return A class name generator function.
     */
    function ClassName(viewName) {
        /**
         * Generates a class name.
         * @param [opt_elementName] The name of the element.
         * @param [opt_modifier] The name of the modifier.
         * @return A class name.
         */
        var fn = function (opt_elementName, opt_modifier) {
            return [
                PREFIX,
                '-',
                viewName,
                'v',
                opt_elementName ? "_" + opt_elementName : '',
                opt_modifier ? "-" + opt_modifier : '',
            ].join('');
        };
        return fn;
    }

    function compose(h1, h2) {
        return function (input) { return h2(h1(input)); };
    }
    function extractValue(ev) {
        return ev.rawValue;
    }
    function applyClass(elem, className, active) {
        if (active) {
            elem.classList.add(className);
        }
        else {
            elem.classList.remove(className);
        }
    }
    function valueToClassName(elem, className) {
        return function (value) {
            applyClass(elem, className, value);
        };
    }
    var className$5 = ClassName('');
    function valueToModifier(elem, modifier) {
        return valueToClassName(elem, className$5(undefined, modifier));
    }
    function bindValue(value, applyValue) {
        value.emitter.on('change', compose(extractValue, applyValue));
        applyValue(value.rawValue);
    }
    function bindValueMap(valueMap, key, applyValue) {
        bindValue(valueMap.value(key), applyValue);
    }
    function bindClassModifier(viewProps, elem) {
        bindValueMap(viewProps, 'disabled', valueToModifier(elem, 'disabled'));
        bindValueMap(viewProps, 'hidden', valueToModifier(elem, 'hidden'));
    }
    function bindDisabled(viewProps, target) {
        bindValueMap(viewProps, 'disabled', function (disabled) {
            target.disabled = disabled;
        });
    }
    function bindTabIndex(viewProps, elem) {
        bindValueMap(viewProps, 'disabled', function (disabled) {
            elem.tabIndex = disabled ? -1 : 0;
        });
    }

    var className$4 = ClassName('ckr');
    var RingView = /** @class */ (function () {
        function RingView(doc, config) {
            this.tickElems_ = [];
            this.labelElems_ = [];
            this.boundsWidth_ = -1;
            this.onShowsTooltipChange_ = this.onShowsTooltipChange_.bind(this);
            this.onValueChange_ = this.onValueChange_.bind(this);
            this.formatters_ = config.formatters;
            this.unit_ = config.unit;
            this.element = doc.createElement('div');
            this.element.classList.add(className$4(), className$4(undefined, "m" + config.seriesId));
            bindClassModifier(config.viewProps, this.element);
            this.value_ = config.value;
            this.value_.emitter.on('change', this.onValueChange_);
            config.showsTooltip.emitter.on('change', this.onShowsTooltipChange_);
            var wrapperElem = doc.createElement('div');
            wrapperElem.classList.add(className$4('w'));
            this.element.appendChild(wrapperElem);
            this.offsetElem_ = doc.createElement('div');
            this.offsetElem_.classList.add(className$4('o'));
            wrapperElem.appendChild(this.offsetElem_);
            this.svgElem_ = doc.createElementNS(SVG_NS, 'svg');
            this.svgElem_.classList.add(className$4('g'));
            this.offsetElem_.appendChild(this.svgElem_);
            this.tooltipElem_ = doc.createElement('div');
            this.tooltipElem_.classList.add(ClassName('tt')(), className$4('tt'));
            this.element.appendChild(this.tooltipElem_);
            this.waitToBeAdded_();
        }
        // Waits to be added to DOM tree to build initial scale elements
        RingView.prototype.waitToBeAdded_ = function () {
            var _this = this;
            var ob = new MutationObserver(function (ml) {
                for (var _i = 0, ml_1 = ml; _i < ml_1.length; _i++) {
                    var m = ml_1[_i];
                    if (m.type !== 'childList') {
                        continue;
                    }
                    m.addedNodes.forEach(function (elem) {
                        if (!elem.contains(_this.element)) {
                            return;
                        }
                        _this.update();
                        ob.disconnect();
                    });
                }
            });
            var doc = this.element.ownerDocument;
            ob.observe(doc.body, {
                attributes: true,
                childList: true,
                subtree: true,
            });
        };
        RingView.prototype.rebuildScaleIfNeeded_ = function (bw) {
            if (this.boundsWidth_ === bw) {
                return;
            }
            this.boundsWidth_ = bw;
            this.tickElems_.forEach(function (elem) {
                removeElement(elem);
            });
            this.tickElems_ = [];
            this.labelElems_.forEach(function (elem) {
                removeElement(elem);
            });
            this.labelElems_ = [];
            var doc = this.element.ownerDocument;
            var tpu = this.unit_.ticks;
            var uw = this.unit_.pixels;
            var halfUnitCount = Math.ceil(bw / 2 / uw) + 1;
            var unitCount = halfUnitCount * 2 + 1;
            var tickCount = unitCount * tpu;
            var tickWidth = uw / tpu;
            for (var i = 0; i < tickCount; i++) {
                var x = i * tickWidth;
                if (i % tpu === 0) {
                    var lineElem = doc.createElementNS(SVG_NS, 'line');
                    lineElem.classList.add(className$4('mjt'));
                    lineElem.setAttributeNS(null, 'x1', String(x));
                    lineElem.setAttributeNS(null, 'y1', '0');
                    lineElem.setAttributeNS(null, 'x2', String(x));
                    lineElem.setAttributeNS(null, 'y2', '2');
                    this.svgElem_.appendChild(lineElem);
                    this.tickElems_.push(lineElem);
                    var labelElem = doc.createElement('div');
                    labelElem.classList.add(className$4('l'));
                    labelElem.style.left = x + "px";
                    this.offsetElem_.appendChild(labelElem);
                    this.labelElems_.push(labelElem);
                }
                else {
                    var lineElem = doc.createElementNS(SVG_NS, 'line');
                    lineElem.classList.add(className$4('mnt'));
                    lineElem.setAttributeNS(null, 'x1', String(x));
                    lineElem.setAttributeNS(null, 'y1', '0');
                    lineElem.setAttributeNS(null, 'x2', String(x));
                    lineElem.setAttributeNS(null, 'y2', '2');
                    this.svgElem_.appendChild(lineElem);
                    this.tickElems_.push(lineElem);
                }
            }
        };
        RingView.prototype.updateScale_ = function (bw) {
            var _this = this;
            var uv = this.unit_.value;
            var uw = this.unit_.pixels;
            var v = this.value_.rawValue;
            var halfUnitCount = Math.ceil(bw / 2 / uw) + 1;
            var ov = v - (v % uv) - uv * halfUnitCount;
            var opacity = function (tv) {
                return (1 -
                    Math.pow(constrainRange(Math.abs(v - tv) / ((bw / 2) * (uv / uw)), 0, 1), 10));
            };
            this.labelElems_.forEach(function (elem, i) {
                var lv = ov + i * uv;
                elem.textContent = _this.formatters_.ring(lv);
                elem.style.opacity = String(opacity(lv));
            });
            var tpu = this.unit_.ticks;
            this.tickElems_.forEach(function (elem, i) {
                var lv = ov + (i / tpu) * uv;
                elem.style.opacity = String(opacity(lv));
            });
        };
        RingView.prototype.update = function () {
            var bw = this.element.getBoundingClientRect().width;
            var uv = this.unit_.value;
            var uw = this.unit_.pixels;
            var v = this.value_.rawValue;
            var halfUnitCount = Math.ceil(bw / 2 / uw) + 1;
            var offsetFromCenter = ((v % uv) + uv * halfUnitCount) * (uw / uv);
            var offset = bw / 2 - offsetFromCenter;
            this.offsetElem_.style.transform = "translateX(" + offset + "px)";
            this.tooltipElem_.textContent = this.formatters_.text(v);
            this.rebuildScaleIfNeeded_(bw);
            this.updateScale_(bw);
        };
        RingView.prototype.onValueChange_ = function () {
            this.update();
        };
        RingView.prototype.onShowsTooltipChange_ = function (ev) {
            if (ev.rawValue) {
                this.element.classList.add(className$4(undefined, 'tt'));
            }
            else {
                this.element.classList.remove(className$4(undefined, 'tt'));
            }
        };
        return RingView;
    }());

    var RingController = /** @class */ (function () {
        function RingController(doc, config) {
            this.ox_ = 0;
            this.ov_ = 0;
            this.onPointerDown_ = this.onPointerDown_.bind(this);
            this.onPointerMove_ = this.onPointerMove_.bind(this);
            this.onPointerUp_ = this.onPointerUp_.bind(this);
            this.value = config.value;
            this.viewProps = config.viewProps;
            this.tooltipEnabled_ = config.tooltipEnabled;
            this.unit_ = config.unit;
            this.showsTooltip_ = new PrimitiveValue(false);
            this.view = new RingView(doc, {
                formatters: config.formatters,
                seriesId: config.seriesId,
                showsTooltip: this.showsTooltip_,
                unit: config.unit,
                value: this.value,
                viewProps: this.viewProps,
            });
            var ptHandler = new PointerHandler(this.view.element);
            ptHandler.emitter.on('down', this.onPointerDown_);
            ptHandler.emitter.on('move', this.onPointerMove_);
            ptHandler.emitter.on('up', this.onPointerUp_);
        }
        RingController.prototype.onPointerDown_ = function (ev) {
            var data = ev.data;
            if (!data.point) {
                return;
            }
            this.ox_ = data.point.x;
            this.ov_ = this.value.rawValue;
            if (this.tooltipEnabled_) {
                this.showsTooltip_.rawValue = true;
            }
        };
        RingController.prototype.onPointerMove_ = function (ev) {
            var data = ev.data;
            if (!data.point) {
                return;
            }
            var dx = data.point.x - this.ox_;
            var uw = this.unit_.pixels;
            var uv = this.unit_.value;
            this.value.rawValue = this.ov_ - (dx / uw) * uv;
        };
        RingController.prototype.onPointerUp_ = function () {
            this.showsTooltip_.rawValue = false;
        };
        return RingController;
    }());

    function createParamFinder(test) {
        return function (params, key) {
            if (!(key in params)) {
                return;
            }
            var value = params[key];
            return test(value) ? value : undefined;
        };
    }
    var findBooleanParam = createParamFinder(function (value) { return typeof value === 'boolean'; });
    var findNumberParam = createParamFinder(function (value) { return typeof value === 'number'; });
    var findStringParam = createParamFinder(function (value) { return typeof value === 'string'; });
    function isObject$1(value) {
        if (value === null) {
            return false;
        }
        return typeof value === 'object';
    }
    var findObjectParam = createParamFinder(isObject$1);

    function isObject(value) {
        if (value === null) {
            return false;
        }
        return typeof value === 'object';
    }
    function findSeries(params, key) {
        var value = findNumberParam(params, key);
        if (value === undefined) {
            return;
        }
        return value === 0 || value === 1 || value === 2 ? value : undefined;
    }
    function createUnit(params, key) {
        var value = findObjectParam(params, key);
        if (!isObject(value)) {
            return;
        }
        var pixels = findNumberParam(value, 'pixels');
        var ticks = findNumberParam(value, 'ticks');
        var unitValue = findNumberParam(value, 'value');
        if (pixels === undefined || ticks === undefined || unitValue === undefined) {
            return;
        }
        return {
            pixels: pixels,
            ticks: ticks,
            value: unitValue,
        };
    }
    function createRingInputParams(params) {
        var plugin = findStringParam(params, 'plugin');
        var view = findStringParam(params, 'view');
        if (plugin !== 'camerakit' || view === undefined) {
            return null;
        }
        return {
            max: findNumberParam(params, 'max'),
            min: findNumberParam(params, 'min'),
            series: findSeries(params, 'series'),
            step: findNumberParam(params, 'step'),
            plugin: plugin,
            unit: createUnit(params, 'unit'),
            view: view,
            wide: findBooleanParam(params, 'wide'),
        };
    }
    function createWheelInputParams(params) {
        var plugin = findStringParam(params, 'plugin');
        var view = findStringParam(params, 'view');
        if (plugin !== 'camerakit' || view === undefined) {
            return null;
        }
        return {
            amount: findNumberParam(params, 'amount'),
            max: findNumberParam(params, 'max'),
            min: findNumberParam(params, 'min'),
            step: findNumberParam(params, 'step'),
            plugin: plugin,
            view: view,
            wide: findBooleanParam(params, 'wide'),
        };
    }

    /**
     * A constraint to combine multiple constraints.
     * @template T The type of the value.
     */
    var CompositeConstraint = /** @class */ (function () {
        function CompositeConstraint(constraints) {
            this.constraints = constraints;
        }
        CompositeConstraint.prototype.constrain = function (value) {
            return this.constraints.reduce(function (result, c) {
                return c.constrain(result);
            }, value);
        };
        return CompositeConstraint;
    }());
    function findConstraint(c, constraintClass) {
        if (c instanceof constraintClass) {
            return c;
        }
        if (c instanceof CompositeConstraint) {
            var result = c.constraints.reduce(function (tmpResult, sc) {
                if (tmpResult) {
                    return tmpResult;
                }
                return sc instanceof constraintClass ? sc : null;
            }, null);
            if (result) {
                return result;
            }
        }
        return null;
    }

    var NumberLiteralNode = /** @class */ (function () {
        function NumberLiteralNode(text) {
            this.text = text;
        }
        NumberLiteralNode.prototype.evaluate = function () {
            return Number(this.text);
        };
        NumberLiteralNode.prototype.toString = function () {
            return this.text;
        };
        return NumberLiteralNode;
    }());
    var BINARY_OPERATION_MAP = {
        '**': function (v1, v2) { return Math.pow(v1, v2); },
        '*': function (v1, v2) { return v1 * v2; },
        '/': function (v1, v2) { return v1 / v2; },
        '%': function (v1, v2) { return v1 % v2; },
        '+': function (v1, v2) { return v1 + v2; },
        '-': function (v1, v2) { return v1 - v2; },
        '<<': function (v1, v2) { return v1 << v2; },
        '>>': function (v1, v2) { return v1 >> v2; },
        '>>>': function (v1, v2) { return v1 >>> v2; },
        '&': function (v1, v2) { return v1 & v2; },
        '^': function (v1, v2) { return v1 ^ v2; },
        '|': function (v1, v2) { return v1 | v2; },
    };
    var BinaryOperationNode = /** @class */ (function () {
        function BinaryOperationNode(operator, left, right) {
            this.left = left;
            this.operator = operator;
            this.right = right;
        }
        BinaryOperationNode.prototype.evaluate = function () {
            var op = BINARY_OPERATION_MAP[this.operator];
            if (!op) {
                throw new Error("unexpected binary operator: '" + this.operator);
            }
            return op(this.left.evaluate(), this.right.evaluate());
        };
        BinaryOperationNode.prototype.toString = function () {
            return [
                'b(',
                this.left.toString(),
                this.operator,
                this.right.toString(),
                ')',
            ].join(' ');
        };
        return BinaryOperationNode;
    }());
    var UNARY_OPERATION_MAP = {
        '+': function (v) { return v; },
        '-': function (v) { return -v; },
        '~': function (v) { return ~v; },
    };
    var UnaryOperationNode = /** @class */ (function () {
        function UnaryOperationNode(operator, expr) {
            this.operator = operator;
            this.expression = expr;
        }
        UnaryOperationNode.prototype.evaluate = function () {
            var op = UNARY_OPERATION_MAP[this.operator];
            if (!op) {
                throw new Error("unexpected unary operator: '" + this.operator);
            }
            return op(this.expression.evaluate());
        };
        UnaryOperationNode.prototype.toString = function () {
            return ['u(', this.operator, this.expression.toString(), ')'].join(' ');
        };
        return UnaryOperationNode;
    }());

    function combineReader(parsers) {
        return function (text, cursor) {
            for (var i = 0; i < parsers.length; i++) {
                var result = parsers[i](text, cursor);
                if (result !== '') {
                    return result;
                }
            }
            return '';
        };
    }
    function readWhitespace(text, cursor) {
        var _a;
        var m = text.substr(cursor).match(/^\s+/);
        return (_a = (m && m[0])) !== null && _a !== void 0 ? _a : '';
    }
    function readNonZeroDigit(text, cursor) {
        var ch = text.substr(cursor, 1);
        return ch.match(/^[1-9]$/) ? ch : '';
    }
    function readDecimalDigits(text, cursor) {
        var _a;
        var m = text.substr(cursor).match(/^[0-9]+/);
        return (_a = (m && m[0])) !== null && _a !== void 0 ? _a : '';
    }
    function readSignedInteger(text, cursor) {
        var ds = readDecimalDigits(text, cursor);
        if (ds !== '') {
            return ds;
        }
        var sign = text.substr(cursor, 1);
        cursor += 1;
        if (sign !== '-' && sign !== '+') {
            return '';
        }
        var sds = readDecimalDigits(text, cursor);
        if (sds === '') {
            return '';
        }
        return sign + sds;
    }
    function readExponentPart(text, cursor) {
        var e = text.substr(cursor, 1);
        cursor += 1;
        if (e.toLowerCase() !== 'e') {
            return '';
        }
        var si = readSignedInteger(text, cursor);
        if (si === '') {
            return '';
        }
        return e + si;
    }
    function readDecimalIntegerLiteral(text, cursor) {
        var ch = text.substr(cursor, 1);
        if (ch === '0') {
            return ch;
        }
        var nzd = readNonZeroDigit(text, cursor);
        cursor += nzd.length;
        if (nzd === '') {
            return '';
        }
        return nzd + readDecimalDigits(text, cursor);
    }
    function readDecimalLiteral1(text, cursor) {
        var dil = readDecimalIntegerLiteral(text, cursor);
        cursor += dil.length;
        if (dil === '') {
            return '';
        }
        var dot = text.substr(cursor, 1);
        cursor += dot.length;
        if (dot !== '.') {
            return '';
        }
        var dds = readDecimalDigits(text, cursor);
        cursor += dds.length;
        return dil + dot + dds + readExponentPart(text, cursor);
    }
    function readDecimalLiteral2(text, cursor) {
        var dot = text.substr(cursor, 1);
        cursor += dot.length;
        if (dot !== '.') {
            return '';
        }
        var dds = readDecimalDigits(text, cursor);
        cursor += dds.length;
        if (dds === '') {
            return '';
        }
        return dot + dds + readExponentPart(text, cursor);
    }
    function readDecimalLiteral3(text, cursor) {
        var dil = readDecimalIntegerLiteral(text, cursor);
        cursor += dil.length;
        if (dil === '') {
            return '';
        }
        return dil + readExponentPart(text, cursor);
    }
    var readDecimalLiteral = combineReader([
        readDecimalLiteral1,
        readDecimalLiteral2,
        readDecimalLiteral3,
    ]);
    function parseBinaryDigits(text, cursor) {
        var _a;
        var m = text.substr(cursor).match(/^[01]+/);
        return (_a = (m && m[0])) !== null && _a !== void 0 ? _a : '';
    }
    function readBinaryIntegerLiteral(text, cursor) {
        var prefix = text.substr(cursor, 2);
        cursor += prefix.length;
        if (prefix.toLowerCase() !== '0b') {
            return '';
        }
        var bds = parseBinaryDigits(text, cursor);
        if (bds === '') {
            return '';
        }
        return prefix + bds;
    }
    function readOctalDigits(text, cursor) {
        var _a;
        var m = text.substr(cursor).match(/^[0-7]+/);
        return (_a = (m && m[0])) !== null && _a !== void 0 ? _a : '';
    }
    function readOctalIntegerLiteral(text, cursor) {
        var prefix = text.substr(cursor, 2);
        cursor += prefix.length;
        if (prefix.toLowerCase() !== '0o') {
            return '';
        }
        var ods = readOctalDigits(text, cursor);
        if (ods === '') {
            return '';
        }
        return prefix + ods;
    }
    function readHexDigits(text, cursor) {
        var _a;
        var m = text.substr(cursor).match(/^[0-9a-f]+/i);
        return (_a = (m && m[0])) !== null && _a !== void 0 ? _a : '';
    }
    function readHexIntegerLiteral(text, cursor) {
        var prefix = text.substr(cursor, 2);
        cursor += prefix.length;
        if (prefix.toLowerCase() !== '0x') {
            return '';
        }
        var hds = readHexDigits(text, cursor);
        if (hds === '') {
            return '';
        }
        return prefix + hds;
    }
    var readNonDecimalIntegerLiteral = combineReader([
        readBinaryIntegerLiteral,
        readOctalIntegerLiteral,
        readHexIntegerLiteral,
    ]);
    var readNumericLiteral = combineReader([
        readNonDecimalIntegerLiteral,
        readDecimalLiteral,
    ]);

    function parseLiteral(text, cursor) {
        var num = readNumericLiteral(text, cursor);
        cursor += num.length;
        if (num === '') {
            return null;
        }
        return {
            evaluable: new NumberLiteralNode(num),
            cursor: cursor,
        };
    }
    function parseParenthesizedExpression(text, cursor) {
        var op = text.substr(cursor, 1);
        cursor += op.length;
        if (op !== '(') {
            return null;
        }
        var expr = parseExpression(text, cursor);
        if (!expr) {
            return null;
        }
        cursor = expr.cursor;
        cursor += readWhitespace(text, cursor).length;
        var cl = text.substr(cursor, 1);
        cursor += cl.length;
        if (cl !== ')') {
            return null;
        }
        return {
            evaluable: expr.evaluable,
            cursor: cursor,
        };
    }
    function parsePrimaryExpression(text, cursor) {
        return (parseLiteral(text, cursor) || parseParenthesizedExpression(text, cursor));
    }
    function parseUnaryExpression(text, cursor) {
        var expr = parsePrimaryExpression(text, cursor);
        if (expr) {
            return expr;
        }
        var op = text.substr(cursor, 1);
        cursor += op.length;
        if (op !== '+' && op !== '-' && op !== '~') {
            return null;
        }
        var num = parseUnaryExpression(text, cursor);
        if (!num) {
            return null;
        }
        cursor = num.cursor;
        return {
            cursor: cursor,
            evaluable: new UnaryOperationNode(op, num.evaluable),
        };
    }
    function readBinaryOperator(ops, text, cursor) {
        cursor += readWhitespace(text, cursor).length;
        var op = ops.filter(function (op) { return text.startsWith(op, cursor); })[0];
        if (!op) {
            return null;
        }
        cursor += op.length;
        cursor += readWhitespace(text, cursor).length;
        return {
            cursor: cursor,
            operator: op,
        };
    }
    function createBinaryOperationExpressionParser(exprParser, ops) {
        return function (text, cursor) {
            var firstExpr = exprParser(text, cursor);
            if (!firstExpr) {
                return null;
            }
            cursor = firstExpr.cursor;
            var expr = firstExpr.evaluable;
            for (;;) {
                var op = readBinaryOperator(ops, text, cursor);
                if (!op) {
                    break;
                }
                cursor = op.cursor;
                var nextExpr = exprParser(text, cursor);
                if (!nextExpr) {
                    return null;
                }
                cursor = nextExpr.cursor;
                expr = new BinaryOperationNode(op.operator, expr, nextExpr.evaluable);
            }
            return expr
                ? {
                    cursor: cursor,
                    evaluable: expr,
                }
                : null;
        };
    }
    var parseBinaryOperationExpression = [
        ['**'],
        ['*', '/', '%'],
        ['+', '-'],
        ['<<', '>>>', '>>'],
        ['&'],
        ['^'],
        ['|'],
    ].reduce(function (parser, ops) {
        return createBinaryOperationExpressionParser(parser, ops);
    }, parseUnaryExpression);
    function parseExpression(text, cursor) {
        cursor += readWhitespace(text, cursor).length;
        return parseBinaryOperationExpression(text, cursor);
    }
    /**
     * Parse ECMAScript expression with numeric literals.
     * https://262.ecma-international.org/
     * @param text The string to be parsed.
     * @return A parsing result, or null if failed.
     */
    function parseEcmaNumberExpression(text) {
        var expr = parseExpression(text, 0);
        if (!expr) {
            return null;
        }
        var cursor = expr.cursor + readWhitespace(text, expr.cursor).length;
        if (cursor !== text.length) {
            return null;
        }
        return expr.evaluable;
    }

    /**
     * @hidden
     */
    function parseNumber(text) {
        var _a;
        var r = parseEcmaNumberExpression(text);
        return (_a = r === null || r === void 0 ? void 0 : r.evaluate()) !== null && _a !== void 0 ? _a : null;
    }
    /**
     * @hidden
     */
    function numberFromUnknown(value) {
        if (typeof value === 'number') {
            return value;
        }
        if (typeof value === 'string') {
            var pv = parseNumber(value);
            if (!isEmpty(pv)) {
                return pv;
            }
        }
        return 0;
    }
    /**
     * @hidden
     */
    function createNumberFormatter(digits) {
        return function (value) {
            return value.toFixed(Math.max(Math.min(digits, 20), 0));
        };
    }

    /**
     * Writes the primitive value.
     * @param target The target to be written.
     * @param value The value to write.
     */
    function writePrimitive(target, value) {
        target.write(value);
    }

    var CREATE_MESSAGE_MAP = {
        alreadydisposed: function () { return 'View has been already disposed'; },
        invalidparams: function (context) { return "Invalid parameters for '" + context.name + "'"; },
        nomatchingcontroller: function (context) {
            return "No matching controller for '" + context.key + "'";
        },
        nomatchingview: function (context) {
            return "No matching view for '" + JSON.stringify(context.params) + "'";
        },
        notbindable: function () { return "Value is not bindable"; },
        propertynotfound: function (context) { return "Property '" + context.name + "' not found"; },
        shouldneverhappen: function () { return 'This error should never happen'; },
    };
    var TpError = /** @class */ (function () {
        function TpError(config) {
            var _a;
            this.message = (_a = CREATE_MESSAGE_MAP[config.type](forceCast(config.context))) !== null && _a !== void 0 ? _a : 'Unexpected error';
            this.name = this.constructor.name;
            this.stack = new Error(this.message).stack;
            this.type = config.type;
        }
        TpError.alreadyDisposed = function () {
            return new TpError({ type: 'alreadydisposed' });
        };
        TpError.notBindable = function () {
            return new TpError({
                type: 'notbindable',
            });
        };
        TpError.propertyNotFound = function (name) {
            return new TpError({
                type: 'propertynotfound',
                context: {
                    name: name,
                },
            });
        };
        TpError.shouldNeverHappen = function () {
            return new TpError({ type: 'shouldneverhappen' });
        };
        return TpError;
    }());
    TpError.prototype = Object.create(Error.prototype);
    TpError.prototype.constructor = TpError;

    /**
     * A list constranit.
     * @template T The type of the value.
     */
    /** @class */ ((function () {
        function ListConstraint(options) {
            this.options = options;
        }
        ListConstraint.prototype.constrain = function (value) {
            var opts = this.options;
            if (opts.length === 0) {
                return value;
            }
            var matched = opts.filter(function (item) {
                return item.value === value;
            }).length > 0;
            return matched ? value : opts[0].value;
        };
        return ListConstraint;
    })());

    /**
     * A number step range constraint.
     */
    var StepConstraint = /** @class */ (function () {
        function StepConstraint(step) {
            this.step = step;
        }
        StepConstraint.prototype.constrain = function (value) {
            var r = value < 0
                ? -Math.round(-value / this.step)
                : Math.round(value / this.step);
            return r * this.step;
        };
        return StepConstraint;
    }());

    var ValueMap = /** @class */ (function () {
        function ValueMap(initialValue) {
            var _this = this;
            this.emitter = new Emitter();
            var keys = Object.keys(initialValue);
            var props = keys.map(function (key) { return new PrimitiveValue(initialValue[key]); });
            props.forEach(function (prop, index) {
                prop.emitter.on('change', function () {
                    _this.emitter.emit('change', {
                        key: keys[index],
                        sender: _this,
                    });
                });
            });
            this.valMap_ = keys.reduce(function (o, key, index) {
                var _a;
                return Object.assign(o, (_a = {},
                    _a[key] = props[index],
                    _a));
            }, {});
        }
        ValueMap.prototype.get = function (key) {
            return this.valMap_[key].rawValue;
        };
        ValueMap.prototype.set = function (key, value) {
            this.valMap_[key].rawValue = value;
        };
        ValueMap.prototype.value = function (key) {
            return this.valMap_[key];
        };
        // TODO: Remove in the next major version
        /** @deprecated Use ValueMap.value.emitter instead. */
        ValueMap.prototype.valueEmitter = function (key) {
            console.warn("ValueMap.valueEmitter is deprecated. Use ValueMap.value.emitter instead.\nThis polyfill will be removed in the next major version.");
            return this.valMap_[key].emitter;
        };
        return ValueMap;
    }());

    function findStep(constraint) {
        var c = constraint ? findConstraint(constraint, StepConstraint) : null;
        if (!c) {
            return null;
        }
        return c.step;
    }
    /**
     * @hidden
     */
    function getSuitableDecimalDigits(constraint, rawValue) {
        var sc = constraint && findConstraint(constraint, StepConstraint);
        if (sc) {
            return getDecimalDigits(sc.step);
        }
        return Math.max(getDecimalDigits(rawValue), 2);
    }
    /**
     * @hidden
     */
    function getBaseStep(constraint) {
        var step = findStep(constraint);
        return step !== null && step !== void 0 ? step : 1;
    }
    /**
     * @hidden
     */
    function getSuitableDraggingScale(constraint, rawValue) {
        var _a;
        var sc = constraint && findConstraint(constraint, StepConstraint);
        var base = Math.abs((_a = sc === null || sc === void 0 ? void 0 : sc.step) !== null && _a !== void 0 ? _a : rawValue);
        return base === 0 ? 0.1 : Math.pow(10, Math.floor(Math.log10(base)) - 1);
    }

    /**
     * A number range constraint.
     */
    var RangeConstraint = /** @class */ (function () {
        function RangeConstraint(config) {
            this.maxValue = config.max;
            this.minValue = config.min;
        }
        RangeConstraint.prototype.constrain = function (value) {
            var result = value;
            if (!isEmpty(this.minValue)) {
                result = Math.max(result, this.minValue);
            }
            if (!isEmpty(this.maxValue)) {
                result = Math.min(result, this.maxValue);
            }
            return result;
        };
        return RangeConstraint;
    }());

    var BoundValue = /** @class */ (function () {
        function BoundValue(initialValue, config) {
            var _a;
            this.constraint_ = config === null || config === void 0 ? void 0 : config.constraint;
            this.equals_ = (_a = config === null || config === void 0 ? void 0 : config.equals) !== null && _a !== void 0 ? _a : (function (v1, v2) { return v1 === v2; });
            this.emitter = new Emitter();
            this.rawValue_ = initialValue;
        }
        Object.defineProperty(BoundValue.prototype, "constraint", {
            get: function () {
                return this.constraint_;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(BoundValue.prototype, "rawValue", {
            get: function () {
                return this.rawValue_;
            },
            set: function (rawValue) {
                var constrainedValue = this.constraint_
                    ? this.constraint_.constrain(rawValue)
                    : rawValue;
                var changed = !this.equals_(this.rawValue_, constrainedValue);
                if (changed) {
                    this.rawValue_ = constrainedValue;
                    this.emitter.emit('change', {
                        rawValue: constrainedValue,
                        sender: this,
                    });
                }
            },
            enumerable: false,
            configurable: true
        });
        return BoundValue;
    }());

    /**
     * @hidden
     */
    function getStepForKey(baseStep, keys) {
        var step = baseStep * (keys.altKey ? 0.1 : 1) * (keys.shiftKey ? 10 : 1);
        if (keys.upKey) {
            return +step;
        }
        else if (keys.downKey) {
            return -step;
        }
        return 0;
    }
    /**
     * @hidden
     */
    function getVerticalStepKeys(ev) {
        return {
            altKey: ev.altKey,
            downKey: ev.key === 'ArrowDown',
            shiftKey: ev.shiftKey,
            upKey: ev.key === 'ArrowUp',
        };
    }
    /**
     * @hidden
     */
    function getHorizontalStepKeys(ev) {
        return {
            altKey: ev.altKey,
            downKey: ev.key === 'ArrowLeft',
            shiftKey: ev.shiftKey,
            upKey: ev.key === 'ArrowRight',
        };
    }

    var className$3 = ClassName('txt');
    var NumberTextView = /** @class */ (function () {
        function NumberTextView(doc, config) {
            this.onChange_ = this.onChange_.bind(this);
            this.props_ = config.props;
            this.props_.emitter.on('change', this.onChange_);
            this.element = doc.createElement('div');
            this.element.classList.add(className$3(), className$3(undefined, 'num'));
            if (config.arrayPosition) {
                this.element.classList.add(className$3(undefined, config.arrayPosition));
            }
            bindClassModifier(config.viewProps, this.element);
            var inputElem = doc.createElement('input');
            inputElem.classList.add(className$3('i'));
            inputElem.type = 'text';
            bindDisabled(config.viewProps, inputElem);
            this.element.appendChild(inputElem);
            this.inputElement = inputElem;
            this.onDraggingChange_ = this.onDraggingChange_.bind(this);
            this.dragging_ = config.dragging;
            this.dragging_.emitter.on('change', this.onDraggingChange_);
            this.element.classList.add(className$3());
            this.inputElement.classList.add(className$3('i'));
            var knobElem = doc.createElement('div');
            knobElem.classList.add(className$3('k'));
            this.element.appendChild(knobElem);
            this.knobElement = knobElem;
            var guideElem = doc.createElementNS(SVG_NS, 'svg');
            guideElem.classList.add(className$3('g'));
            this.knobElement.appendChild(guideElem);
            var bodyElem = doc.createElementNS(SVG_NS, 'path');
            bodyElem.classList.add(className$3('gb'));
            guideElem.appendChild(bodyElem);
            this.guideBodyElem_ = bodyElem;
            var headElem = doc.createElementNS(SVG_NS, 'path');
            headElem.classList.add(className$3('gh'));
            guideElem.appendChild(headElem);
            this.guideHeadElem_ = headElem;
            var tooltipElem = doc.createElement('div');
            tooltipElem.classList.add(ClassName('tt')());
            this.knobElement.appendChild(tooltipElem);
            this.tooltipElem_ = tooltipElem;
            config.value.emitter.on('change', this.onChange_);
            this.value = config.value;
            this.refresh();
        }
        NumberTextView.prototype.onDraggingChange_ = function (ev) {
            if (ev.rawValue === null) {
                this.element.classList.remove(className$3(undefined, 'drg'));
                return;
            }
            this.element.classList.add(className$3(undefined, 'drg'));
            var x = ev.rawValue / this.props_.get('draggingScale');
            var aox = x + (x > 0 ? -1 : x < 0 ? +1 : 0);
            var adx = constrainRange(-aox, -4, +4);
            this.guideHeadElem_.setAttributeNS(null, 'd', ["M " + (aox + adx) + ",0 L" + aox + ",4 L" + (aox + adx) + ",8", "M " + x + ",-1 L" + x + ",9"].join(' '));
            this.guideBodyElem_.setAttributeNS(null, 'd', "M 0,4 L" + x + ",4");
            var formatter = this.props_.get('formatter');
            this.tooltipElem_.textContent = formatter(this.value.rawValue);
            this.tooltipElem_.style.left = x + "px";
        };
        NumberTextView.prototype.refresh = function () {
            var formatter = this.props_.get('formatter');
            this.inputElement.value = formatter(this.value.rawValue);
        };
        NumberTextView.prototype.onChange_ = function () {
            this.refresh();
        };
        return NumberTextView;
    }());

    /**
     * @hidden
     */
    var NumberTextController = /** @class */ (function () {
        function NumberTextController(doc, config) {
            this.originRawValue_ = 0;
            this.onInputChange_ = this.onInputChange_.bind(this);
            this.onInputKeyDown_ = this.onInputKeyDown_.bind(this);
            this.onPointerDown_ = this.onPointerDown_.bind(this);
            this.onPointerMove_ = this.onPointerMove_.bind(this);
            this.onPointerUp_ = this.onPointerUp_.bind(this);
            this.baseStep_ = config.baseStep;
            this.parser_ = config.parser;
            this.props = config.props;
            this.value = config.value;
            this.viewProps = config.viewProps;
            this.dragging_ = new BoundValue(null);
            this.view = new NumberTextView(doc, {
                arrayPosition: config.arrayPosition,
                dragging: this.dragging_,
                props: this.props,
                value: this.value,
                viewProps: this.viewProps,
            });
            this.view.inputElement.addEventListener('change', this.onInputChange_);
            this.view.inputElement.addEventListener('keydown', this.onInputKeyDown_);
            var ph = new PointerHandler(this.view.knobElement);
            ph.emitter.on('down', this.onPointerDown_);
            ph.emitter.on('move', this.onPointerMove_);
            ph.emitter.on('up', this.onPointerUp_);
        }
        NumberTextController.prototype.onInputChange_ = function (e) {
            var inputElem = forceCast(e.currentTarget);
            var value = inputElem.value;
            var parsedValue = this.parser_(value);
            if (!isEmpty(parsedValue)) {
                this.value.rawValue = parsedValue;
            }
            this.view.refresh();
        };
        NumberTextController.prototype.onInputKeyDown_ = function (e) {
            var step = getStepForKey(this.baseStep_, getVerticalStepKeys(e));
            if (step !== 0) {
                this.value.rawValue += step;
            }
        };
        NumberTextController.prototype.onPointerDown_ = function () {
            this.originRawValue_ = this.value.rawValue;
            this.dragging_.rawValue = 0;
        };
        NumberTextController.prototype.onPointerMove_ = function (ev) {
            if (!ev.data.point) {
                return;
            }
            var dx = ev.data.point.x - ev.data.bounds.width / 2;
            this.value.rawValue =
                this.originRawValue_ + dx * this.props.get('draggingScale');
            this.dragging_.rawValue = this.value.rawValue - this.originRawValue_;
        };
        NumberTextController.prototype.onPointerUp_ = function () {
            this.dragging_.rawValue = null;
        };
        return NumberTextController;
    }());

    var className$2 = ClassName('sldtxt');
    /**
     * @hidden
     */
    var SliderTextView = /** @class */ (function () {
        function SliderTextView(doc, config) {
            this.element = doc.createElement('div');
            this.element.classList.add(className$2());
            var sliderElem = doc.createElement('div');
            sliderElem.classList.add(className$2('s'));
            this.sliderView_ = config.sliderView;
            sliderElem.appendChild(this.sliderView_.element);
            this.element.appendChild(sliderElem);
            var textElem = doc.createElement('div');
            textElem.classList.add(className$2('t'));
            this.textView_ = config.textView;
            textElem.appendChild(this.textView_.element);
            this.element.appendChild(textElem);
        }
        return SliderTextView;
    }());

    var className$1 = ClassName('sld');
    /**
     * @hidden
     */
    var SliderView = /** @class */ (function () {
        function SliderView(doc, config) {
            this.onChange_ = this.onChange_.bind(this);
            this.props_ = config.props;
            this.props_.emitter.on('change', this.onChange_);
            this.element = doc.createElement('div');
            this.element.classList.add(className$1());
            bindClassModifier(config.viewProps, this.element);
            var trackElem = doc.createElement('div');
            trackElem.classList.add(className$1('t'));
            bindTabIndex(config.viewProps, trackElem);
            this.element.appendChild(trackElem);
            this.trackElement = trackElem;
            var knobElem = doc.createElement('div');
            knobElem.classList.add(className$1('k'));
            this.trackElement.appendChild(knobElem);
            this.knobElement = knobElem;
            config.value.emitter.on('change', this.onChange_);
            this.value = config.value;
            this.update_();
        }
        SliderView.prototype.update_ = function () {
            var p = constrainRange(mapRange(this.value.rawValue, this.props_.get('minValue'), this.props_.get('maxValue'), 0, 100), 0, 100);
            this.knobElement.style.width = p + "%";
        };
        SliderView.prototype.onChange_ = function () {
            this.update_();
        };
        return SliderView;
    }());

    /**
     * @hidden
     */
    var SliderController = /** @class */ (function () {
        function SliderController(doc, config) {
            this.onKeyDown_ = this.onKeyDown_.bind(this);
            this.onPoint_ = this.onPoint_.bind(this);
            this.baseStep_ = config.baseStep;
            this.value = config.value;
            this.viewProps = config.viewProps;
            this.props = config.props;
            this.view = new SliderView(doc, {
                props: this.props,
                value: this.value,
                viewProps: this.viewProps,
            });
            this.ptHandler_ = new PointerHandler(this.view.trackElement);
            this.ptHandler_.emitter.on('down', this.onPoint_);
            this.ptHandler_.emitter.on('move', this.onPoint_);
            this.ptHandler_.emitter.on('up', this.onPoint_);
            this.view.trackElement.addEventListener('keydown', this.onKeyDown_);
        }
        SliderController.prototype.handlePointerEvent_ = function (d) {
            if (!d.point) {
                return;
            }
            this.value.rawValue = mapRange(constrainRange(d.point.x, 0, d.bounds.width), 0, d.bounds.width, this.props.get('minValue'), this.props.get('maxValue'));
        };
        SliderController.prototype.onPoint_ = function (ev) {
            this.handlePointerEvent_(ev.data);
        };
        SliderController.prototype.onKeyDown_ = function (ev) {
            this.value.rawValue += getStepForKey(this.baseStep_, getHorizontalStepKeys(ev));
        };
        return SliderController;
    }());

    /**
     * @hidden
     */
    /** @class */ ((function () {
        function SliderTextController(doc, config) {
            this.value = config.value;
            this.viewProps = config.viewProps;
            this.sliderC_ = new SliderController(doc, {
                baseStep: config.baseStep,
                props: config.sliderProps,
                value: config.value,
                viewProps: this.viewProps,
            });
            this.textC_ = new NumberTextController(doc, {
                baseStep: config.baseStep,
                parser: config.parser,
                props: config.textProps,
                value: config.value,
                viewProps: config.viewProps,
            });
            this.view = new SliderTextView(doc, {
                sliderView: this.sliderC_.view,
                textView: this.textC_.view,
            });
        }
        Object.defineProperty(SliderTextController.prototype, "sliderController", {
            get: function () {
                return this.sliderC_;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(SliderTextController.prototype, "textController", {
            get: function () {
                return this.textC_;
            },
            enumerable: false,
            configurable: true
        });
        return SliderTextController;
    })());

    /**
     * Tries to create a step constraint.
     * @param params The input parameters object.
     * @return A constraint or null if not found.
     */
    function createStepConstraint(params) {
        if ('step' in params && !isEmpty(params.step)) {
            return new StepConstraint(params.step);
        }
        return null;
    }
    /**
     * Tries to create a range constraint.
     * @param params The input parameters object.
     * @return A constraint or null if not found.
     */
    function createRangeConstraint(params) {
        if (('max' in params && !isEmpty(params.max)) ||
            ('min' in params && !isEmpty(params.min))) {
            return new RangeConstraint({
                max: params.max,
                min: params.min,
            });
        }
        return null;
    }

    var className = ClassName('ckrtxt');
    var RingTextView = /** @class */ (function () {
        function RingTextView(doc, config) {
            this.element = doc.createElement('div');
            this.element.classList.add(className());
            var ringElem = doc.createElement('div');
            ringElem.classList.add(className('r'));
            ringElem.appendChild(config.ringView.element);
            this.element.appendChild(ringElem);
            var textElem = doc.createElement('div');
            textElem.classList.add(className('t'));
            textElem.appendChild(config.textView.element);
            this.element.appendChild(textElem);
        }
        return RingTextView;
    }());

    var RingTextController = /** @class */ (function () {
        function RingTextController(doc, config) {
            this.value = config.value;
            this.viewProps = config.viewProps;
            this.rc_ = new RingController(doc, {
                formatters: config.formatters,
                seriesId: config.seriesId,
                tooltipEnabled: false,
                unit: config.ringUnit,
                value: this.value,
                viewProps: this.viewProps,
            });
            this.tc_ = new NumberTextController(doc, {
                baseStep: config.baseStep,
                parser: config.parser,
                props: new ValueMap({
                    draggingScale: config.draggingScale,
                    formatter: config.formatters.text,
                }),
                value: this.value,
                viewProps: this.viewProps,
            });
            this.view = new RingTextView(doc, {
                ringView: this.rc_.view,
                textView: this.tc_.view,
            });
        }
        return RingTextController;
    }());

    function getRingSeries(series) {
        return series !== undefined ? String(series) : '0';
    }
    function createRingFormatter(ringUnit) {
        var f = createNumberFormatter(getDecimalDigits(ringUnit.value));
        return function (value) {
            var text = f(value);
            var ch = text.substr(0, 1);
            var hasSign = ch === '-' || ch === '+';
            return text + (hasSign ? ' ' : '');
        };
    }
    function createConstraint(params) {
        var constraints = [];
        var cr = createRangeConstraint(params);
        if (cr) {
            constraints.push(cr);
        }
        var cs = createStepConstraint(params);
        if (cs) {
            constraints.push(cs);
        }
        return new CompositeConstraint(constraints);
    }
    var RingInputPlugin = {
        id: 'input-camerakit-ring',
        css: '.tp-ckrv_w{-webkit-appearance:none;-moz-appearance:none;appearance:none;background-color:transparent;border-width:0;font-family:inherit;font-size:inherit;font-weight:inherit;margin:0;outline:none;padding:0}.tp-ckrv_w{background-color:var(--in-bg);border-radius:var(--elm-br);box-sizing:border-box;color:var(--in-fg);font-family:inherit;height:var(--bld-h);line-height:var(--bld-h);min-width:0;width:100%}.tp-ckrv_w:hover{background-color:var(--in-bg-h)}.tp-ckrv_w:focus{background-color:var(--in-bg-f)}.tp-ckrv_w:active{background-color:var(--in-bg-a)}.tp-ckrv_w:disabled{opacity:0.5}.tp-ckrv{position:relative}.tp-ckrv.tp-v-disabled{opacity:0.5}.tp-ckrv::before{background-color:var(--separator-color);bottom:0;content:\'\';height:2px;left:0;margin:auto;position:absolute;right:0;top:0}.tp-ckrv::after{background-color:var(--in-fg);content:\'\';left:0;margin:auto;position:absolute;right:0}.tp-ckrv_w{cursor:pointer;height:calc(var(--unit-size));overflow:hidden}.tp-ckrv_o{height:100%;left:0;position:relative}.tp-ckrv_g{display:block;height:2px;overflow:visible;position:absolute;width:100%}.tp-ckrv_mjt{stroke:var(--in-fg);stroke-width:2;transform-origin:bottom}.tp-ckrv_mnt{stroke:var(--in-fg);stroke-width:1;transform-origin:bottom}.tp-ckrv_l{color:var(--in-fg);font-size:0.9em;pointer-events:none;position:absolute;top:0;transform:translateX(-50%);white-space:pre}.tp-ckrv_tt{left:50%;top:-4px;visibility:hidden}.tp-ckrv.tp-ckrv-tt .tp-ckrv_tt{visibility:visible}.tp-ckrv.tp-ckrv-m0::before{display:none}.tp-ckrv.tp-ckrv-m0::after{bottom:0;height:6px;width:2px}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_g{bottom:0;margin:auto;transform:translateY(3px);top:0}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_mjt{transform:scaleY(3)}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_mnt{transform:scaleY(2)}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_l{line-height:8px;transform:translateX(-50%) scale(0.8)}.tp-ckrv.tp-ckrv-m1::before{display:none}.tp-ckrv.tp-ckrv-m1::after{bottom:0;height:8px;width:2px}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_g{bottom:0}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_mjt{stroke-width:1;transform:scaleY(4)}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_mnt{transform:scaleY(2)}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_l{line-height:12px}.tp-ckrv.tp-ckrv-m2::before{transform:translateY(3px)}.tp-ckrv.tp-ckrv-m2::after{border-radius:2px;bottom:1px;height:4px;width:4px}.tp-ckrv.tp-ckrv-m2 .tp-ckrv_g{display:none}.tp-ckrv.tp-ckrv-m2 .tp-ckrv_l{line-height:12px;transform:translateX(-50%)}.tp-ckrv.tp-ckrv-mw::before,.tp-ckrv.tp-ckrv-mw::after{display:none}.tp-ckrv.tp-ckrv-mw .tp-ckrv_g{bottom:0;opacity:0.2}.tp-ckrv.tp-ckrv-mw .tp-ckrv_mjt,.tp-ckrv.tp-ckrv-mw .tp-ckrv_mnt{stroke:var(--in-fg);stroke-width:2;transform:scaleY(10)}.tp-ckrv.tp-ckrv-mw .tp-ckrv_l{display:none}.tp-ckrtxtv{display:flex}.tp-ckrtxtv_r{flex:2}.tp-ckrtxtv_t{flex:1;margin-left:4px}',
        accept: function (exValue, params) {
            if (typeof exValue !== 'number') {
                return null;
            }
            var p = createRingInputParams(forceCast(params));
            if (!p) {
                return null;
            }
            if (p.view !== 'ring') {
                return null;
            }
            return exValue;
        },
        binding: {
            reader: function (_args) { return numberFromUnknown; },
            constraint: function (args) { return createConstraint(args.params); },
            writer: function (_args) { return writePrimitive; },
        },
        controller: function (args) {
            var _a, _b, _c;
            var params = createRingInputParams(forceCast(args.params));
            if (!params) {
                throw TpError.shouldNeverHappen();
            }
            var ringUnit = (_a = params.unit) !== null && _a !== void 0 ? _a : {
                ticks: 5,
                pixels: 40,
                value: 10,
            };
            var c = args.constraint;
            var formatters = {
                ring: createRingFormatter(ringUnit),
                text: createNumberFormatter(getSuitableDecimalDigits(c, args.initialValue)),
            };
            if (params.wide) {
                return new RingController(args.document, {
                    formatters: formatters,
                    seriesId: (_b = getRingSeries(params.series)) !== null && _b !== void 0 ? _b : '0',
                    tooltipEnabled: true,
                    unit: ringUnit,
                    value: args.value,
                    viewProps: args.viewProps,
                });
            }
            return new RingTextController(args.document, {
                baseStep: getBaseStep(c),
                draggingScale: getSuitableDraggingScale(c, args.initialValue),
                formatters: formatters,
                parser: parseNumber,
                seriesId: (_c = getRingSeries(params.series)) !== null && _c !== void 0 ? _c : '0',
                ringUnit: ringUnit,
                value: args.value,
                viewProps: args.viewProps,
            });
        },
    };

    var WheelInputPlugin = {
        id: 'input-camerakit-wheel',
        css: '.tp-ckrv_w{-webkit-appearance:none;-moz-appearance:none;appearance:none;background-color:transparent;border-width:0;font-family:inherit;font-size:inherit;font-weight:inherit;margin:0;outline:none;padding:0}.tp-ckrv_w{background-color:var(--in-bg);border-radius:var(--elm-br);box-sizing:border-box;color:var(--in-fg);font-family:inherit;height:var(--bld-h);line-height:var(--bld-h);min-width:0;width:100%}.tp-ckrv_w:hover{background-color:var(--in-bg-h)}.tp-ckrv_w:focus{background-color:var(--in-bg-f)}.tp-ckrv_w:active{background-color:var(--in-bg-a)}.tp-ckrv_w:disabled{opacity:0.5}.tp-ckrv{position:relative}.tp-ckrv.tp-v-disabled{opacity:0.5}.tp-ckrv::before{background-color:var(--separator-color);bottom:0;content:\'\';height:2px;left:0;margin:auto;position:absolute;right:0;top:0}.tp-ckrv::after{background-color:var(--in-fg);content:\'\';left:0;margin:auto;position:absolute;right:0}.tp-ckrv_w{cursor:pointer;height:calc(var(--unit-size));overflow:hidden}.tp-ckrv_o{height:100%;left:0;position:relative}.tp-ckrv_g{display:block;height:2px;overflow:visible;position:absolute;width:100%}.tp-ckrv_mjt{stroke:var(--in-fg);stroke-width:2;transform-origin:bottom}.tp-ckrv_mnt{stroke:var(--in-fg);stroke-width:1;transform-origin:bottom}.tp-ckrv_l{color:var(--in-fg);font-size:0.9em;pointer-events:none;position:absolute;top:0;transform:translateX(-50%);white-space:pre}.tp-ckrv_tt{left:50%;top:-4px;visibility:hidden}.tp-ckrv.tp-ckrv-tt .tp-ckrv_tt{visibility:visible}.tp-ckrv.tp-ckrv-m0::before{display:none}.tp-ckrv.tp-ckrv-m0::after{bottom:0;height:6px;width:2px}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_g{bottom:0;margin:auto;transform:translateY(3px);top:0}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_mjt{transform:scaleY(3)}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_mnt{transform:scaleY(2)}.tp-ckrv.tp-ckrv-m0 .tp-ckrv_l{line-height:8px;transform:translateX(-50%) scale(0.8)}.tp-ckrv.tp-ckrv-m1::before{display:none}.tp-ckrv.tp-ckrv-m1::after{bottom:0;height:8px;width:2px}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_g{bottom:0}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_mjt{stroke-width:1;transform:scaleY(4)}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_mnt{transform:scaleY(2)}.tp-ckrv.tp-ckrv-m1 .tp-ckrv_l{line-height:12px}.tp-ckrv.tp-ckrv-m2::before{transform:translateY(3px)}.tp-ckrv.tp-ckrv-m2::after{border-radius:2px;bottom:1px;height:4px;width:4px}.tp-ckrv.tp-ckrv-m2 .tp-ckrv_g{display:none}.tp-ckrv.tp-ckrv-m2 .tp-ckrv_l{line-height:12px;transform:translateX(-50%)}.tp-ckrv.tp-ckrv-mw::before,.tp-ckrv.tp-ckrv-mw::after{display:none}.tp-ckrv.tp-ckrv-mw .tp-ckrv_g{bottom:0;opacity:0.2}.tp-ckrv.tp-ckrv-mw .tp-ckrv_mjt,.tp-ckrv.tp-ckrv-mw .tp-ckrv_mnt{stroke:var(--in-fg);stroke-width:2;transform:scaleY(10)}.tp-ckrv.tp-ckrv-mw .tp-ckrv_l{display:none}.tp-ckrtxtv{display:flex}.tp-ckrtxtv_r{flex:2}.tp-ckrtxtv_t{flex:1;margin-left:4px}',
        accept: function (exValue, params) {
            if (typeof exValue !== 'number') {
                return null;
            }
            var p = createWheelInputParams(forceCast(params));
            if (!p) {
                return null;
            }
            if (p.view !== 'wheel') {
                return null;
            }
            return exValue;
        },
        binding: {
            reader: function (_args) { return numberFromUnknown; },
            constraint: function (args) { return createConstraint(args.params); },
            writer: function (_args) { return writePrimitive; },
        },
        controller: function (args) {
            var _a, _b;
            var params = createWheelInputParams(forceCast(args.params));
            if (!params) {
                throw TpError.shouldNeverHappen();
            }
            var c = args.constraint;
            var draggingScale = getSuitableDraggingScale(c, args.initialValue);
            var formatters = {
                ring: createNumberFormatter(0),
                text: createNumberFormatter(getSuitableDecimalDigits(c, args.initialValue)),
            };
            if (params.wide) {
                return new RingController(args.document, {
                    formatters: formatters,
                    seriesId: 'w',
                    tooltipEnabled: true,
                    unit: {
                        ticks: 10,
                        pixels: 40,
                        value: ((_a = params.amount) !== null && _a !== void 0 ? _a : draggingScale) * 40,
                    },
                    value: args.value,
                    viewProps: args.viewProps,
                });
            }
            return new RingTextController(args.document, {
                baseStep: getBaseStep(c),
                draggingScale: draggingScale,
                formatters: formatters,
                parser: parseNumber,
                seriesId: 'w',
                ringUnit: {
                    ticks: 10,
                    pixels: 40,
                    value: ((_b = params.amount) !== null && _b !== void 0 ? _b : draggingScale) * 40,
                },
                value: args.value,
                viewProps: args.viewProps,
            });
        },
    };

    {
        Tweakpane__default['default'].registerPlugin({
            type: 'input',
            plugin: RingInputPlugin,
        });
        Tweakpane__default['default'].registerPlugin({
            type: 'input',
            plugin: WheelInputPlugin,
        });
    }

})));
